"""
Analytical approximation.
"""

import numpy as np
from numpy import cos, sin, pi
import matplotlib.pyplot as plt
import lmfit
import phoebe
from phoebe.atmospheres import limbdark
from phoebe.algorithms import interp_nDgrid
try:
    from phoebe.utils import transit
    occult = transit.occultquad
except ImportError:
    print("Cannot import transit models")

cc = phoebe.constants.cc
GG = phoebe.constants.GG
Mjup = phoebe.constants.Mjup
Msol = phoebe.constants.Msol

def lc(pars, phase, data=None, eps=None,
       prim_eclipse='nonlin', sec_eclipse='uniform',
       passband='KEPLER.V', atm1='kurucz', atm2='blackbody'):
    """
    Analytical phae curve representation.
    
    If ``data=None``, the model is returned, otherwise ther residuals.
    
    Parameters:
    
        - f0: continuum flux
        - Fn: night side flux
        - incl: orbital inclination angle [deg]
        - Ag: geometrical albedo
        - Mp: companion mass [Msol]
        - Mstar: primary mass [Msol]
        - period: orbital period [d]
        - a_rel: relative semi-major axis (i.e. in units of primary radius)
        - r_rel: relative companion radius (i.e. in units of primary radius)
        - u: linear limb darkening parameter
        - y: gravity darkening parameter
        - alpha_d: Doppler boosting factor
        - c_i: Claret limb darkening parameters
    """
    f0 = pars['f0'].value
    Fn = pars['Fn'].value
    i = pars['incl'].value
    A_g = pars['Ag'].value
    Mp = pars['Mp'].value
    Mstar = pars['Mstar'].value
    period = pars['period'].value
    a_rel = pars['a_rel'].value
    r_rel = pars['r_rel'].value
    u = pars['u'].value
    y = pars['y'].value
    teff1 = pars['teff1'].value
    teff2 = pars['teff2'].value
    logg1 = pars['logg1'].value
    logg2 = pars['logg2'].value
    abun1 = pars['abun1'].value
    abun2 = pars['abun2'].value
    alpha_d = pars['alpha_d'].value
    c0 = pars['c0'].value
    c1 = pars['c1'].value
    c2 = pars['c2'].value
    c3 = pars['c3'].value
    
    Cd = pars['Cd'].value
    Ce = pars['Ce'].value
    
    # Derive limbdarkening parameters
    limbdark.interp_ld_coeffs
    
    
    i = i/180.*pi
    period = period*24*3600.
    Mp = Mp*Mjup
    Mstar = Mstar*Msol
    Fecl = A_g*(r_rel/a_rel)**2
    Fn = Fn*1e-6
    A_p = Fecl - Fn
    
    sin_i = sin(i)
    cos_z = -sin_i*cos(2*pi*phase)
    sin_z = np.sqrt( 1 - cos_z)**2
    z = np.arccos(cos_z)
        
    alpha_1 = 25.*u/(24.*(15+u)) * (y+2.)/(y+1.)
    alpha_2 = 3.*(15.+u) / (20.*(3.-u)) * (y+1)
    K1 = Cd*(2*pi*GG/period)**(1./3.) * Mp*sin_i/Mstar**(2./3.)
    
    f1 = 3.*alpha_1 / a_rel * (5*sin_i**2 -4)/sin_i
    f2 = 5.*alpha_1 / a_rel * sin_i
    A_e = Ce*alpha_2 * Mp/Mstar / a_rel**3 * sin_i**2
    
    #F_p = A_p * (sin_z + (pi-z)*cos_z) / pi
    
    z = np.abs(2*pi*phase-pi)*sin(i)
    F_p = A_p/ pi * (sin(z) + (pi-z)*cos(z)) + Fn
    
    F_d = alpha_d * K1 / cc * sin(2*pi*phase)
    F_e = -A_e * ( cos(2*pi*2*phase) + f1*cos(2*pi*phase) + f2*cos(2*pi*3*phase))
    star = f0 + F_d + F_e
    
    # Add Primary and Secondary eclipse
    planet = F_p
    rel_flux = None
    
    if prim_eclipse is not None or sec_eclipse is not None:
        z = a_rel*np.sqrt(cos(i)**2 + sin(i)**2*sin(2*pi*phase)**2)
        
        p = r_rel
        cn = [c0, c1, c2, c3]
        
        rel_flux = np.zeros_like(phase)
        back = cos(2*pi*phase) < 0
            
        if prim_eclipse is not Noen:
            star[-back] = (getattr(transit,'occult'+prim_eclipse)(z[-back],p,cn))*star[-back]
        
        if sec_eclipse:
            z = z/p
            p = 1./p
            planet[back] = (1-getattr(transit,'occult'+sec_eclipse)(z[back],p))*F_p[back]
    
    model = star + planet
    
    if data is None:
        return model, f0, F_p, F_d, F_e, rel_flux
    if eps is None:
        return (model - data)/data
    
   
    
    return (model-data)/eps

if __name__=='__main__':

    axis_values, pixelgrid, labels = limbdark._prepare_grid('KEPLER.V', 'beaming/table.fits',
                                data_columns=['alpha_b', 'a1', 'imu'], log_columns=['imu'],
                                nointerp_columns=['alpha_b', 'a1', 'imu'])

    teff = 5850.
    logg = 4.4
    abun = -0.1
    vgamma = 0.0

    alpha_b, u, Imu = interp_nDgrid.interpolate(np.array([[teff, logg, abun, vgamma]]).T, axis_values, pixelgrid)


    pars = lmfit.Parameters()
    pars.add('f0', value=1, min=0, max=180, vary=False)
    pars.add('i', value=83.8646, min=0, max=180, vary=False)
    pars.add('A_p', value=4.77, min=0, max=180, vary=False)
    pars.add('Mp', value=1.33, min=0, max=180, vary=False)
    pars.add('Mstar', value=0.98, min=0, max=180, vary=False)
    pars.add('period', value=2.4706132, min=0, max=180, vary=False)
    pars.add('a_rel', value=7.8957, min=0, max=180, vary=False)
    pars.add('u', value=u, min=0, max=180, vary=False)
    pars.add('r_rel', value=0.125106, min=0, max=180, vary=False)
    pars.add('beta', value=0.6, min=0, max=180, vary=False)
    pars.add('alpha_d', value=alpha_b, min=0, max=180, vary=False)
    pars.add('c0', value=0.3529, min=0, max=1, vary=False)
    pars.add('c1', value=0.2635, min=0, max=1, vary=False)


    phase = np.linspace(0,1,1000)
    curve, f0, F_p, F_d, F_e, F_ecl = residual(pars, phase)

    plt.plot(phase, (f0+F_p)*1e6,label='Planet')
    plt.plot(phase, (f0+F_d)*1e6,label='Doppler')
    plt.plot(phase, (f0+F_e)*1e6,label='Ellipsoidal')
    plt.plot(phase, curve*1e6, 'k-', lw=2, label='total')
    plt.legend(loc='best').get_frame().set_alpha(0.5)


    print (curve*1e6).max()
    print (curve*1e6).min()
    print (curve*1e6).ptp()
    plt.show()
    