"""
Calculate with units and constants

.. autosummary::

    conversions.convert
    conversions.Unit

Section 1 List of constants
===========================

+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Name   | Value                     | Description                 |Units      | Reference                                                |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
| cc    | 299792458.0               | speed of light              |m/s        |                                                          |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|cc_air | cc/1.0002773              | speed of light in air       |m/s        |   T=15C, standard atmospheric pressure                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Msol   | 1.988547e30               | solar mass                  |kg         |   Harmanec & Prsa 2011                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Rsol   | 6.95508e8                 | solar radius                |m          |   Harmanec & Prsa 2011                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Lsol   | 3.846e26                  | solar luminosity            |W          |   Harmanec & Prsa 2011                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Tsol   | 5779.5747                 | solar effective temperature |K          |   Harmanec & Prsa 2011                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|agesol | 1.44221e+17               | solar age                   |s          |   Bahcall 2005                                           |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Mearth | 5.9742e24                 | earth mass                  |kg         |   google                                                 |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Rearth | 6371.e3                   | earth radius                |m          |   wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Mjup   | 1.8986e27                 | Jupiter mass                |kg         |   wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Rjup   | 6.9911e7                  | Jupiter radius              |m          |   mesa                                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|Mlun   | 7.346e22                  | Lunar mass                  |kg         |   wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|au     | 149597870700.0            | astronomical unit           |m          |   Harmanec & Prsa 2011                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|pc     | 3.085677581503e+16        | parsec                      |m          |   Harmanec & Prsa 2011                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|ly     | 9.460730472580800e+15     | light year                  |m          |                                                          |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|hh     | 6.6260689633e-34          | Planck constant             |J/s        |                                                          |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|hhbar  | 1.05457162853e-34         | reduced Planck constant     |J/s        |                                                          |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|kB     | 1.380650424e-23           | Boltzmann constant          |J/K        |                                                          |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|NA     | 6.0221417930e23           | Avogadro constant           |1/mol      |                                                          |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|sigma  | 5.67040040e-8             | Stefan-Boltzmann constant   |W/m2/K4    |   Harmanec & Prsa 2011                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|GG     | 6.67384e-11               | gravitational constant      |m3/kg/s2   |   Harmanec & Prsa 2011                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|GGMsol | 1.32712442099e20          | grav. constant x solar mass |m3/s2      |   Harmanec & Prsa 2011                                   |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|RR     | 8.31447215                | (ideal) gas constant        |J/K/mol    |                                                          |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|aa     | 7.5657e-16                | radiation constant          |J/m3/K4    |                                                          |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|a0     | 52.9177e-12               | Bohr radius of hydrogen     |m          |                                                          |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|ke     | 8.9875517873681764e9      | Coulomb constant            |Nm2/C2     |   Wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|eps0   | 8.854187817620e-12        | Electric constant           |F/m        |   Wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|mu0    | 1.2566370614e-6           | Magnetic constant           |N/A2       |   Wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|alpha  | 137.036                   | Fine structure constant     |none       |    http://pntpm3.ulb.ac.be/private/divers.htm#constants  |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|me     | 9.1093829140e-31          | Electron mass               |kg         |   Wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|qe     |-1.60217656535e-19         | Electron charge             |C          |   Wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|mp     | 1.67262177774e-27         | Proton mass                 |kg         |   Wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+
|qp     | 1.60217656535e-19         | Proton charge               |C          |   Wikipedia                                              |
+-------+---------------------------+-----------------------------+-----------+----------------------------------------------------------+

Section 2 List of recognised units
===================================

**Units of absorbed dose**

+----------------+-------------------------------------------+
|Gy              | gray                                      |
+----------------+-------------------------------------------+

**Units of acceleration**

+----------------+-------------------------------------------+
|Gal             | Galileo                                   |
+----------------+-------------------------------------------+

**Units of ampl_flux**

+----------------+-------------------------------------------+
|Amag            | amplitude in magnitude                    |
+----------------+-------------------------------------------+
|ampl            | fractional amplitude                      |
+----------------+-------------------------------------------+
|pph             | amplitude in parts per hundred            |
+----------------+-------------------------------------------+
|ppm             | amplitude in parts per million            |
+----------------+-------------------------------------------+
|ppt             | amplitude in parts per thousand           |
+----------------+-------------------------------------------+

**Units of angle**

+----------------+-------------------------------------------+
|am              | arcminute                                 |
+----------------+-------------------------------------------+
|as              | arcsecond                                 |
+----------------+-------------------------------------------+
|cy              | cycle                                     |
+----------------+-------------------------------------------+
|deg             | degree                                    |
+----------------+-------------------------------------------+
|rad             | radian                                    |
+----------------+-------------------------------------------+
|sr              | sterradian                                |
+----------------+-------------------------------------------+

**Units of area**

+----------------+-------------------------------------------+
|a               | are                                       |
+----------------+-------------------------------------------+
|ac              | acre (international)                      |
+----------------+-------------------------------------------+
|b               | barn                                      |
+----------------+-------------------------------------------+

**Units of calendar date**

+----------------+-------------------------------------------+
|CD              | calender day                              |
+----------------+-------------------------------------------+

**Units of catalytic activity**

+----------------+-------------------------------------------+
|kat             | katal                                     |
+----------------+-------------------------------------------+

**Units of coordinate**

+----------------+-------------------------------------------+
|complex_coord   | <own unit>                                |
+----------------+-------------------------------------------+
|deg_coord       | degrees                                   |
+----------------+-------------------------------------------+
|ecliptic        | ecliptic                                  |
+----------------+-------------------------------------------+
|equatorial      | equatorial                                |
+----------------+-------------------------------------------+
|galactic        | galactic                                  |
+----------------+-------------------------------------------+
|rad_coord       | radians                                   |
+----------------+-------------------------------------------+

**Units of currency**

+----------------+-------------------------------------------+
|EUR             | EURO                                      |
+----------------+-------------------------------------------+

**Units of date**

+----------------+-------------------------------------------+
|MJD             | modified Julian day                       |
+----------------+-------------------------------------------+

**Units of dose equivalent**

+----------------+-------------------------------------------+
|Sv              | sievert                                   |
+----------------+-------------------------------------------+
|rem             | Roentgen equivelant in man                |
+----------------+-------------------------------------------+

**Units of dynamic viscosity**

+----------------+-------------------------------------------+
|P               | poise                                     |
+----------------+-------------------------------------------+

**Units of electric capacitance**

+----------------+-------------------------------------------+
|F               | Farad                                     |
+----------------+-------------------------------------------+

**Units of electric charge**

+----------------+-------------------------------------------+
|C               | Coulomb                                   |
+----------------+-------------------------------------------+

**Units of electric conductance**

+----------------+-------------------------------------------+
|S               | Siemens                                   |
+----------------+-------------------------------------------+

**Units of electric current**

+----------------+-------------------------------------------+
|A               | Ampere                                    |
+----------------+-------------------------------------------+
|Bi              | Biot                                      |
+----------------+-------------------------------------------+
|Gi              | Ampere                                    |
+----------------+-------------------------------------------+

**Units of electric potential difference**

+----------------+-------------------------------------------+
|V               | Volt                                      |
+----------------+-------------------------------------------+

**Units of electric resistance**

+----------------+-------------------------------------------+
|O               | Ohm                                       |
+----------------+-------------------------------------------+

**Units of energy**

+----------------+-------------------------------------------+
|Cal             | large calorie (international table)       |
+----------------+-------------------------------------------+
|J               | Joule                                     |
+----------------+-------------------------------------------+
|cal             | small calorie (international table)       |
+----------------+-------------------------------------------+
|eV              | electron volt                             |
+----------------+-------------------------------------------+
|erg             | ergon                                     |
+----------------+-------------------------------------------+
|foe             | (ten to the) fifty one ergs               |
+----------------+-------------------------------------------+

**Units of energy/power**

+----------------+-------------------------------------------+
|Lsol            | Solar luminosity                          |
+----------------+-------------------------------------------+

**Units of flux density**

+----------------+-------------------------------------------+
|ABmag           | AB magnitude                              |
+----------------+-------------------------------------------+
|Jy              | Jansky                                    |
+----------------+-------------------------------------------+
|STmag           | ST magnitude                              |
+----------------+-------------------------------------------+
|mag             | magnitude                                 |
+----------------+-------------------------------------------+
|vegamag         | Vega magnitude                            |
+----------------+-------------------------------------------+

**Units of flux_ratio**

+----------------+-------------------------------------------+
|flux_ratio      | flux ratio                                |
+----------------+-------------------------------------------+
|mag_color       | color                                     |
+----------------+-------------------------------------------+

**Units of force**

+----------------+-------------------------------------------+
|N               | Newton                                    |
+----------------+-------------------------------------------+
|dyn             | dyne                                      |
+----------------+-------------------------------------------+

**Units of frequency**

+----------------+-------------------------------------------+
|Bq              | Becquerel                                 |
+----------------+-------------------------------------------+
|Ci              | Curie                                     |
+----------------+-------------------------------------------+
|hz              | Hertz                                     |
+----------------+-------------------------------------------+
|rpm             | revolutions per minute                    |
+----------------+-------------------------------------------+

**Units of illuminance**

+----------------+-------------------------------------------+
|lx              | lux                                       |
+----------------+-------------------------------------------+
|ph              | phot                                      |
+----------------+-------------------------------------------+
|sb              | stilb                                     |
+----------------+-------------------------------------------+

**Units of inductance**

+----------------+-------------------------------------------+
|H               | Henry                                     |
+----------------+-------------------------------------------+

**Units of kynamatic viscosity**

+----------------+-------------------------------------------+
|St              | stokes                                    |
+----------------+-------------------------------------------+

**Units of length**

+----------------+-------------------------------------------+
|AA              | angstrom                                  |
+----------------+-------------------------------------------+
|AU              | astronomical unit                         |
+----------------+-------------------------------------------+
|Rearth          | Earth radius                              |
+----------------+-------------------------------------------+
|Rjup            | Jupiter radius                            |
+----------------+-------------------------------------------+
|Rsol            | Solar radius                              |
+----------------+-------------------------------------------+
|USft            | foot (US)                                 |
+----------------+-------------------------------------------+
|USmi            | mile (US)                                 |
+----------------+-------------------------------------------+
|bs              | beard second                              |
+----------------+-------------------------------------------+
|ch              | chain                                     |
+----------------+-------------------------------------------+
|ell             | ell                                       |
+----------------+-------------------------------------------+
|fathom          | fathom                                    |
+----------------+-------------------------------------------+
|ft              | foot (international)                      |
+----------------+-------------------------------------------+
|fur             | furlong                                   |
+----------------+-------------------------------------------+
|in              | inch (international)                      |
+----------------+-------------------------------------------+
|ly              | light year                                |
+----------------+-------------------------------------------+
|m               | meter                                     |
+----------------+-------------------------------------------+
|mi              | mile (international)                      |
+----------------+-------------------------------------------+
|nami            | nautical mile                             |
+----------------+-------------------------------------------+
|pc              | parsec                                    |
+----------------+-------------------------------------------+
|perch           | pole                                      |
+----------------+-------------------------------------------+
|pole            | perch                                     |
+----------------+-------------------------------------------+
|potrzebie       | potrzebie                                 |
+----------------+-------------------------------------------+
|rd              | rod                                       |
+----------------+-------------------------------------------+
|smoot           | smooth                                    |
+----------------+-------------------------------------------+
|yd              | yard (international)                      |
+----------------+-------------------------------------------+

**Units of luminous flux**

+----------------+-------------------------------------------+
|lm              | lumen                                     |
+----------------+-------------------------------------------+

**Units of m/s**

+----------------+-------------------------------------------+
|cc              | Speed of light                            |
+----------------+-------------------------------------------+

**Units of magnetic field strength**

+----------------+-------------------------------------------+
|G               | Gauss                                     |
+----------------+-------------------------------------------+
|T               | Tesla                                     |
+----------------+-------------------------------------------+

**Units of magnetic flux**

+----------------+-------------------------------------------+
|Mx              | Maxwell                                   |
+----------------+-------------------------------------------+
|Wb              | Weber                                     |
+----------------+-------------------------------------------+

**Units of magnetizing field**

+----------------+-------------------------------------------+
|Oe              | Oersted                                   |
+----------------+-------------------------------------------+

**Units of mass**

+----------------+-------------------------------------------+
|Mearth          | Earth mass                                |
+----------------+-------------------------------------------+
|Mjup            | Jupiter mass                              |
+----------------+-------------------------------------------+
|Mlun            | Lunar mass                                |
+----------------+-------------------------------------------+
|Msol            | Solar mass                                |
+----------------+-------------------------------------------+
|carat           | carat                                     |
+----------------+-------------------------------------------+
|firkin          | firkin                                    |
+----------------+-------------------------------------------+
|g               | gram                                      |
+----------------+-------------------------------------------+
|gr              | gram                                      |
+----------------+-------------------------------------------+
|lb              | pound                                     |
+----------------+-------------------------------------------+
|ounce           | ounce                                     |
+----------------+-------------------------------------------+
|st              | stone                                     |
+----------------+-------------------------------------------+
|ton             | gram                                      |
+----------------+-------------------------------------------+
|u               | atomic mass                               |
+----------------+-------------------------------------------+

**Units of molarmass**

+----------------+-------------------------------------------+
|mol             | molar mass                                |
+----------------+-------------------------------------------+

**Units of power**

+----------------+-------------------------------------------+
|W               | Watt                                      |
+----------------+-------------------------------------------+
|dp              | Donkeypower                               |
+----------------+-------------------------------------------+
|hp              | Horsepower                                |
+----------------+-------------------------------------------+

**Units of pressure**

+----------------+-------------------------------------------+
|Pa              | Pascal                                    |
+----------------+-------------------------------------------+
|at              | atmosphere (technical)                    |
+----------------+-------------------------------------------+
|atm             | atmosphere (standard)                     |
+----------------+-------------------------------------------+
|ba              | barye                                     |
+----------------+-------------------------------------------+
|bar             | baros                                     |
+----------------+-------------------------------------------+
|mmHg            | millimeter of mercury                     |
+----------------+-------------------------------------------+
|psi             | pound per square inch                     |
+----------------+-------------------------------------------+
|torr            | Torricelli                                |
+----------------+-------------------------------------------+

**Units of roentgen**

+----------------+-------------------------------------------+
|R               | Roentgen                                  |
+----------------+-------------------------------------------+

**Units of second**

+----------------+-------------------------------------------+
|fortnight       | fortnight                                 |
+----------------+-------------------------------------------+

**Units of sidereal time**

+----------------+-------------------------------------------+
|sidereal        | sidereal day                              |
+----------------+-------------------------------------------+

**Units of temperature**

+----------------+-------------------------------------------+
|Cel             | Celcius                                   |
+----------------+-------------------------------------------+
|Far             | Fahrenheit                                |
+----------------+-------------------------------------------+
|K               | Kelvin                                    |
+----------------+-------------------------------------------+
|Tsol            | Solar temperature                         |
+----------------+-------------------------------------------+

**Units of time**

+----------------+-------------------------------------------+
|JD              | Julian day                                |
+----------------+-------------------------------------------+
|cr              | century                                   |
+----------------+-------------------------------------------+
|d               | day                                       |
+----------------+-------------------------------------------+
|h               | hour                                      |
+----------------+-------------------------------------------+
|j               | jiffy                                     |
+----------------+-------------------------------------------+
|min             | minute                                    |
+----------------+-------------------------------------------+
|mo              | month                                     |
+----------------+-------------------------------------------+
|s               | second                                    |
+----------------+-------------------------------------------+
|wk              | week                                      |
+----------------+-------------------------------------------+
|yr              | year                                      |
+----------------+-------------------------------------------+

**Units of velocity**

+----------------+-------------------------------------------+
|knot            | nautical mile per hour                    |
+----------------+-------------------------------------------+
|mph             | miles per hour                            |
+----------------+-------------------------------------------+

**Units of volume**

+----------------+-------------------------------------------+
|USgal           | gallon (US)                               |
+----------------+-------------------------------------------+
|USgi            | gill (Canadian and UK imperial)           |
+----------------+-------------------------------------------+
|bbl             | barrel                                    |
+----------------+-------------------------------------------+
|bu              | bushel                                    |
+----------------+-------------------------------------------+
|gal             | gallon (Canadian and UK imperial)         |
+----------------+-------------------------------------------+
|gi              | gill (Canadian and UK imperial)           |
+----------------+-------------------------------------------+
|l               | liter                                     |
+----------------+-------------------------------------------+
|ngogn           | 1000 cubic potrzebies                     |
+----------------+-------------------------------------------+
"""